// Solovay Strassen test with Chaitin Schwartz construction.
// g++ -O2 CSSS_Test3.cpp -o CSSS_Test3 -lgmp -lgmpxx

#include <iostream>
#include <fstream>
#include <vector>
#include <string>
#include <cassert>
#include <cstdlib>
#include <limits>
#include <cmath>

#include <gmpxx.h>

#define JUMP_BITS 1
#define BIT_FLIP 1

using namespace std;

struct EOF_Error 
{ 
    int e;
    EOF_Error(int e2) { e = e2; }
};

class BitSeq  // sequence of bits from a raw binary file
{
    private:
    ifstream in;
	
    char byte;	// current byte read from stream
    int offset;  	// last bit 7..0 of byte used

    public:
    BitSeq(const char *fname) : in(fname) { offset=-1; }
    ~BitSeq() { in.close(); }
    int eof() { return (in.fail() && in.eof()); }

    char getbit()
    {
#ifdef BIT_FLIP
	if (offset==-1) { in.get(byte); byte = ~byte;  offset=7; }
#else
	if (offset==-1) { in.get(byte);  offset=7; }
#endif
	return byte & 1<<offset-- ? 1 : 0;
    }

    unsigned char reverse(unsigned char b)
    {
        b = (b & 0xF0) >> 4 | (b & 0x0F) << 4;
        b = (b & 0xCC) >> 2 | (b & 0x33) << 2;
        b = (b & 0xAA) >> 1 | (b & 0x55) << 1;
        return b;
    }

    void jumpbits(unsigned long long cnt)
    {
        offset=-1;
        in.seekg(cnt/8);
        getbits(cnt%8);
	if (in.eof()) throw EOF_Error(1); 

    }

    mpz_class getbits(unsigned long long cnt)
    {
	mpz_class t(0);

        mpz_class tmp;

        unsigned long long total=cnt;

	if (offset < cnt)   // can we finish off current byte?
	{// note: if offset == -1 then t will still be 0			
	    char temp = byte & (1<<offset+1)-1;
            t = reverse((unsigned char)temp)>>(7-offset);
	    cnt -= offset+1;
	    offset=-1; 
	}

	// fast version (by bytes)
	//
        if (cnt>=8) offset=-1;  // set once
	while (cnt>=8) 
	{
	    in.get(byte);
#ifdef BIT_FLIP
            byte = ~byte;
#endif
	    if (in.eof()) throw EOF_Error(1);

            mpz_ui_pow_ui(tmp.get_mpz_t(),2, total-cnt);
	    t += reverse((unsigned char) byte)*tmp; 
	    cnt -= 8;
	}

	// slow version (continue by bits)
	//
	while (cnt)
        {
            mpz_ui_pow_ui(tmp.get_mpz_t(),2, total-cnt);
            t+=getbit()*tmp;
            cnt--;
        }
	if (in.eof()) throw EOF_Error(2);
	return t; 
    }
};

BitSeq *inbits;
  
bool solovaytest(mpz_class &n, mpz_class &b)
{
	mpz_class nm1(n), x(1L), y;
	nm1 = nm1 - 1L;
  
	int jacobi = mpz_jacobi(b.get_mpz_t(), n.get_mpz_t());
    
	if (jacobi==0) return true;

	mpz_class nm1d2(nm1);
	nm1d2 = nm1d2 / 2;
	// test a^(n-1/2) != jacobi (mod n)
	//
	mpz_powm(y.get_mpz_t(), b.get_mpz_t(), nm1d2.get_mpz_t(), n.get_mpz_t());

	if ((jacobi==-1 && y!=nm1) || (jacobi==1 && y!=1)) return true;
	return false;
}
    
int main(int argn, char *argc[])
{
	// A sequence of big int pairs is read from keyboard
  	//
  	// Arguments: input bit filename and number of tests plus
  	// an optional read offset from bit file
  	assert(argn==2 || argn==3);
    
        int rounds=-1;
        if(argn==3) rounds=stoi(argc[2]); //How many Carmichael number will be test
        inbits = new BitSeq(argc[1]);
  	int c = 0; //store the value of c. c comes from c-random.

  	cout << "Salovay-Strassen tests with Chaitin Schwartz construction results table " << endl;

  	mpz_class J; //sotre the Carmichael number

  	try
	{

		unsigned long long num_carmichael=0, tot_count = 0;

		while (true)
		{
#if JUMP_BITS
                        //re-start with some bits jumped
                        inbits->jumpbits(tot_count);

#endif
                        if(rounds==0) break;
			cin >> J;  if (J==0) break;
			cout<<"Carmichael number " << J <<": ";
			num_carmichael++;

			int j = mpz_sizeinbase(J.get_mpz_t(),2);// binary legnth of J
                        mpz_class J_minus_1(J-1L);
                        int bits_used = mpz_sizeinbase(J_minus_1.get_mpz_t(),2);// approximation of binary bits used for 1 bit of J-1 based number.

			c = j-1; //c = floor(log2(J)), since J is not power of 2 (Carmichael numbers are not even), c=j-1
			int readLength = j*(j+2*c); // read length according to Chaitin Schwartz theorem
                         
			mpz_class s(0);//Interpret string as an integer

			s = inbits->getbits(readLength);

			mpz_class d(0);//store base J-1 bit

			bool composite=0;
			unsigned long count = 0;
                        
			//Calculate J-1 representation of s (d_kd_{k-1}...d_0)
			//then do SS(J,1+d_{k-1}), SS(J, 1+d_{k-2}) ...  until one test says composite
                        mpz_class end(J-1L);
                        while(!composite)
                        {
                            mpz_class tmp(s);
			    while(tmp>end)//Stop when i = k (do not calculate the right most d)
			    {
			        d = tmp%(J-1L);
				tmp = tmp/(J-1L);
                            }
                            d = d+1L;

			    composite = solovaytest(J, d);
                            end = end*(J-1L);
			    count++;
      
      			}
                        cout<<"Trial number = "<<count<<" Bits used = "<<bits_used*count<<endl;
                        tot_count += bits_used*count;
                        if(rounds>0)rounds--;

    		}
                cout<<"Total bits used = "<<tot_count<<endl;

	}
  	catch (EOF_Error e){
    		cerr << "Read past end of file!\n";
  	}

}

