// Solovay Strassen test with Chaitin Schwartz construction.

#include <iostream>
#include <fstream>
#include <vector>
#include <string>
#include <cassert>
#include <cstdlib>
#include <limits>
#include <cmath>

#include <gmpxx.h>

#define JUMP_BITS 1
#define BIT_FLIP 1

using namespace std;

struct EOF_Error 
{ 
    int e;
    EOF_Error(int e2) { e = e2; }
};

class BitSeq  // sequence of bits from a raw binary file
{
    private:
    ifstream in;
	
    char byte;	// current byte read from stream
    int offset;  	// last bit 7..0 of byte used

    public:
    BitSeq(const char *fname) : in(fname) { offset=-1; }
    ~BitSeq() { in.close(); }
    int eof() { return (in.fail() && in.eof()); }

    char getbit()
    {
#ifdef BIT_FLIP
	if (offset==-1) { in.get(byte); byte = ~byte;  offset=7; }
#else
	if (offset==-1) { in.get(byte);  offset=7; }
#endif
	return byte & 1<<offset-- ? 1 : 0;
    }

//Nan's code start
    unsigned char reverse(unsigned char b)
    {
        b = (b & 0xF0) >> 4 | (b & 0x0F) << 4;
        b = (b & 0xCC) >> 2 | (b & 0x33) << 2;
        b = (b & 0xAA) >> 1 | (b & 0x55) << 1;
        return b;
    }

    void jumpbits(unsigned long long cnt)
    {
        offset=-1;
        in.seekg(cnt/8);
        getbits(cnt%8);
	if (in.eof()) throw EOF_Error(1); 

    }

    mpz_class getbits(unsigned long long cnt)
    {
	mpz_class t(0);

        mpz_class tmp;

        unsigned long long total=cnt;

	if (offset < cnt)   // can we finish off current byte?
	{// note: if offset == -1 then t will still be 0			
	    char temp = byte & (1<<offset+1)-1;
            t = reverse((unsigned char)temp)>>(7-offset);
	    cnt -= offset+1;
	    offset=-1; 
	}

	// fast version (by bytes)
	//
        if (cnt>=8) offset=-1;  // set once
	while (cnt>=8) 
	{
	    in.get(byte);
#ifdef BIT_FLIP
            byte = ~byte;
#endif
	    if (in.eof()) throw EOF_Error(1);

            mpz_ui_pow_ui(tmp.get_mpz_t(),2, total-cnt);
	    t += reverse((unsigned char) byte)*tmp; 
	    cnt -= 8;
	}
//Nan's code end
	// slow version (continue by bits)
	//
	while (cnt)
        {
            mpz_ui_pow_ui(tmp.get_mpz_t(),2, total-cnt);
            t+=getbit()*tmp;
            cnt--;
        }
	if (in.eof()) throw EOF_Error(2);
	return t; 
    }
};

bool *ss_table;

BitSeq *inbits;
  
bool solovaytest(mpz_class &n, mpz_class &b)
{
    mpz_class nm1(n), x(1L), y;
    nm1 = nm1 - 1L;

    int jacobi = mpz_jacobi(b.get_mpz_t(), n.get_mpz_t());
    
    if (jacobi==0) return true;

    mpz_class nm1d2(nm1);
    nm1d2 = nm1d2 / 2;
    // test a^(n-1/2) != jacobi (mod n)
    //
    mpz_powm(y.get_mpz_t(), b.get_mpz_t(), nm1d2.get_mpz_t(), n.get_mpz_t());

    if ((jacobi==-1 && y!=nm1) || (jacobi==1 && y!=1)) return true;
    return false;
}

bool* cal_ss_table(mpz_class &n)
{
    bool* ss_table = new bool[n.get_ui()-1];
    for(unsigned long i=0;i<n.get_ui()-1;i++)
     {
         mpz_class a(i+1);
         ss_table[i]=solovaytest(n,a);
     }
     return ss_table;
}

int main(int argn, char *argc[])
{
    // A sequence of big int pairs is read from keyboard
    //
    // Arguments: input bit filename and number of tests plus
    // an optional read offset from bit file
    assert(argn==2);

    // BitSeq inbits(argc[1]);     // random bit source filename
    inbits = new BitSeq(argc[1]);     // random bit source filename

    int c = 0; //store the value of c. c comes from c-random.

    cout << "Salovay-Strassen tests with Chaitin Schwartz construction results table " << endl;

    mpz_class J; //sotre the Carmichael number

    while(true)//loop to go through all test numbers
    {
        unsigned long long count = 0;
        cin >> J;
        if(J==0) return 0;
        ss_table = cal_ss_table(J);//Build table for look up for all possible test witnesses w.r.t. J
        int j = mpz_sizeinbase(J.get_mpz_t(),2);// binary legnth of J
        mpz_class J_minus_1(J-1L);
        int bits_used = mpz_sizeinbase(J_minus_1.get_mpz_t(),2);// approximation of binary bits used for 1 bit of J-1 based number.

        c = j-1; //c = floor(log2(J)), since J is not power of 2, c=j-1
        int readLength = j*(j+2*c); // read length according to Chaitin Schwartz theorem
        unsigned long long offset_file = 0;
        while(readLength-offset_file)//
        {
            inbits->jumpbits(offset_file);
            try
            {
                while (true)// loop to go through a random sample with a fixed test number
                {
                    bool flag = true;
     
                    mpz_class s(0);//Interpret string as an integer

                    s = inbits->getbits(readLength);

                    mpz_class d(0);//store base J-1 bit
                        
                    //Calculate J-1 representation of s (d_kd_{k-1}...d_0)
                    //then do SS(J,1+d_0), SS(J, 1+d_1) ...  until one test says composite
                    while(s>=(J-1L))//Stop when i = k (do not calculate the left most d)
                    {
                        d = s%(J-1L);
                        s = s/(J-1L);

                        //composite = solovaytest(J, d);

                        if (ss_table[d.get_ui()])//not add 1 to d, since the result is stored in index 0 when d=1
                        {
                            flag=false;
                            break;
                        }
                    }
                    if(flag) count++;
                }

            }
            catch (EOF_Error e){
            // Initialize the enviroment for next number (exception happens
            // when not enough bits to read from the random sample)
                delete inbits;
                inbits = new BitSeq(argc[1]);
            }
            //cout<<"Violation count = "<<count<<endl;
            offset_file++;
        }
        cout<<"Violation count = " << count << endl;
        delete ss_table;
        bool *ss_table;
    }
}
